#ifdef __APPLE__
/* Availability Header */
#include <Availability.h>
#endif /* __APPLE__ */

/* Mach Headers */
#include <mach/mach_types.h>
#include <mach/machine.h>
#include <mach/processor.h>
#include <mach/processor_info.h>

/* Pexpert Headers */
#include <pexpert/pexpert.h>

/* String Headers */
#include <string.h>

/* IOKit Headers */
#include <IOKit/IOLib.h>
#include <IOKit/IOService.h>
#include <IOKit/IOUserClient.h>
#include <IOKit/IOBufferMemoryDescriptor.h>
#include <IOKit/IODeviceMemory.h>
//#include <IOKit/IOStatistics.h>
#include <IOKit/IORegistryEntry.h>
#include <IOKit/IOLocks.h>
//#include <IOKit/IOStatistics.h>

#ifndef __has_include
#define __has_include(...) 0
#endif /* __has_include */

#if __has_include(<IOKit/IOSharedDataQueue.h>)
#include <IOKit/IOSharedDataQueue.h>

#ifndef USE_OLD_IODATAQUEUE
#define USE_OLD_IODATAQUEUE 0
#endif /*  USE_OLD_IODATAQUEUE */
#else /* !__has_include(<IOKit/IOSharedDataQueue.h>) */
#include <IOKit/IODataQueue.h>

#ifndef USE_OLD_IODATAQUEUE
#define USE_OLD_IODATAQUEUE 1
#endif /*  USE_OLD_IODATAQUEUE */
#endif /* __has_include(<IOKit/IOSharedDataQueue.h>) */

#ifndef BUFSIZE
#define BUFSIZE PAGE_SIZE /* Buffer Size In Bytes => Page Size (Defaults To 4 Kilo Bytes). */
#endif /* BUFSIZE */

#ifndef MAXENTRIES
#define MAXENTRIES ((PAGE_SIZE) / 4) /* Maximum User Clients => 1024 Entries. */
#endif /* MAXENTRIES */

#ifndef MAXUSERS
#define MAXUSERS ((MAXENTRIES) / 8) /* Maximum Active User Clients => 128 Users. */
#endif /* MAXUSERS */

#ifndef kMethodObjectUserClient
#define kMethodObjectUserClient ((IOService *)0)
#endif /* kMethodObjectUserClient */

#if defined(__GNUC__) || defined(__clang__)
#ifndef APPLE_KEXT_OVERRIDE
#define APPLE_KEXT_OVERRIDE override
#endif /* __GNUC__ || __clang__ */

#ifndef ATTRIBUTE_MACRO
#define ATTRIBUTE_MACRO __attribute__((format(printf, 3, 4)))
#endif /* ATTRIBUTE_MACRO */
#else /* !__GNUC__ && !__clang__ */
#ifndef APPLE_KEXT_OVERRIDE
#define APPLE_KEXT_OVERRIDE /* No Override Macro */
#endif /* APPLE_KEXT_OVERRIDE */

#ifndef ATTRIBUTE_MACRO
#define ATTRIBUTE_MACRO /* No Attribute Macro */
#endif /* ATTRIBUTE_MACRO */
#endif /* __GNUC__ || __clang__ */

#ifdef __LP64__
/* 64-Bit Size Macros */
#ifndef IOPhysicalAddress
#define IOPhysicalAddress UInt64
#endif /* IOPhysicalAddress */

#ifndef IOPhysicalLength
#define IOPhysicalLength UInt64
#endif /* IOPhysicalLength */

#ifndef IOByteCount
#define IOByteCount UInt32
#endif /* IOByteCount */
#else /* __LP32__ */
/* 32-Bit Size Macros */
#ifndef IOPhysicalAddress
#define IOPhysicalAddress UInt32
#endif /* IOPhysicalAddress */

#ifndef IOPhysicalLength
#define IOPhysicalLength UInt32
#endif /* IOPhysicalLength */

#ifndef IOByteCount
#define IOByteCount UInt32
#endif /* IOByteCount */
#endif /* __LPXX__ */

/* Read Register Macros */
#ifndef rdmsr
#define rdmsr(msr,lo,hi) __asm__ volatile("rdmsr" : "=a" (lo), "=d" (hi) : "c" (msr))
#endif /* rdmsr */

#ifndef rdpmc
#define rdpmc(counter,lo,hi) __asm__ volatile("rdpmc" : "=a" (lo), "=d" (hi) : "c" (counter))
#endif /* rdpmc */

#ifndef rdtsc
#define rdtsc(lo,hi) __asm__ volatile("lfence; rdtsc" : "=a" (lo), "=d" (hi))
#endif /* rdtsc */

#ifndef rdtsc_nofence
#define rdtsc_nofence(lo,hi) __asm__ volatile("rdtsc" : "=a" (lo), "=d" (hi))
#endif /* rdtsc_nofence */

/* Write Register Macros */
#ifndef wrmsr
#define wrmsr(msr,lo,hi) __asm__ volatile("wrmsr" :: "c" (msr), "a" (lo), "d" (hi))
#endif /* wrmsr */

#ifndef wrpmc
#define wrpmc(counter,lo,hi) wrmsr((0x309 + (counter)), (lo), (hi))
#endif /* wrpmc */

#ifndef write_tsc
#define write_tsc(lo,hi) wrmsr(0x10, (lo), (hi))
#endif /* write_tsc */

/* Unused Argument Macro */
#ifndef __unused
#define __unused __attribute__((unused))
#endif /* __unused */

/* AnVMSRAction  */
typedef enum _AnVMSRAction
{
    AnVMSRActionMethodRDMSR = 0,
    AnVMSRActionMethodWRMSR = 1,
    AnVMSRActionMethodRDPMC = 2,
    AnVMSRActionMethodWRPMC = 3,
    AnVMSRActionMethodRDTSC = 4,
    AnVMSRActionMethodWRTSC = 5,
    AnVMSRNumMethods /* 6 Methods */
} AnVMSRAction;

typedef struct _msr_inout
{
    /* 32-Bit Action Selection */
	UInt32 action;

    /* 32-Bit MSR Value */
    UInt32 msr;

    /* 64-Bit Parameter Value */
    UInt64 param;
} msr_inout;

class AnVMSRUserClient;

class AnVMSR : public IOService
{
    OSDeclareDefaultStructors(AnVMSR)

private:
    /* Private */
    IOByteCount mPrefPanelMemoryBufSize = 0; /* User Client Memory Buffer Size */
    IOPhysicalLength mClientCount = 0; /* User Client Count */
    IOPhysicalAddress mPrefPanelMemoryBuf[AnVMSRNumMethods] = { 0, 0, 0, 0, 0, 0 }; /* User Client Memory Buffer */
    UInt64 mIORegID = 0; /* Registry Entry ID */
    AnVMSRUserClient *mClientPtr[MAXUSERS+1] = { NULL }; /* User Client Pointer */
    OSDictionary *mPropTable = NULL; /* IORegistry Resource Properties */
    IOService *mResService = NULL; /* IOResource IOService */
    IORegistryEntry *mIORegEntry = NULL; /* IORegistry Entry */
    bool mErrFlag = false; /* Error Flag */

public:
    /* IOKit Methods */
    virtual bool init(void) APPLE_KEXT_OVERRIDE;
	virtual bool init(OSDictionary *dictionary = 0) APPLE_KEXT_OVERRIDE;
	virtual void free(void) APPLE_KEXT_OVERRIDE;
	virtual bool start(IOService *provider) APPLE_KEXT_OVERRIDE;
	virtual void stop(IOService *provider) APPLE_KEXT_OVERRIDE;

    /* Setter/Getter Methods */
    inline IOByteCount getMemorySize(void)
    {
        return (mPrefPanelMemoryBufSize);
    }

    inline void setMemorySize(IOByteCount Size)
    {
        mPrefPanelMemoryBufSize = Size;
    }

    inline IOPhysicalAddress *getMemoryBuf(void)
    {
        return ((IOPhysicalAddress *)mPrefPanelMemoryBuf);
    }

    inline void setMemoryBuf(IOPhysicalAddress *Addr)
    {
        memcpy(mPrefPanelMemoryBuf, Addr, sizeof(mPrefPanelMemoryBuf));
    }

    /* Read Methods */
    virtual UInt64 a_rdmsr(UInt32 msr);             /* RDMSR */
    virtual UInt64 a_rdpmc(UInt32 pmc);             /* RDPMC */
    virtual UInt64 a_rdtsc(void);                   /* RDTSC */

    /* Write Methods */
    virtual void a_wrmsr(UInt32 msr, UInt64 value); /* WRMSR */
    virtual void a_wrpmc(UInt32 pmc, UInt64 value); /* WRPMC */
    virtual void a_wrtsc(UInt64 value);             /* WRTSC */

    /* Standard User Client Methods */
    virtual IOReturn newUserClient(task_t owningTask, void *securityID,
                                   UInt32 type, OSDictionary *properties,
                                   IOUserClient **handler) APPLE_KEXT_OVERRIDE;
    virtual IOReturn newUserClient(task_t owningTask, void *securityID,
                                   UInt32 type, IOUserClient **handler) APPLE_KEXT_OVERRIDE;

    /* Custom User Client Methods */
    virtual IOReturn runAction(AnVMSRAction action, UInt32 *outSize, void **outData, void *extraArg);
    virtual bool closeChild(AnVMSRUserClient *ptr);
    virtual void setErr(bool set);

    /* Virtual Padding Functions For AnVMSR */
    OSMetaClassDeclareReservedUnused(AnVMSR, 0);
    OSMetaClassDeclareReservedUnused(AnVMSR, 1);
    OSMetaClassDeclareReservedUnused(AnVMSR, 2);
    OSMetaClassDeclareReservedUnused(AnVMSR, 3);
    OSMetaClassDeclareReservedUnused(AnVMSR, 4);
    OSMetaClassDeclareReservedUnused(AnVMSR, 5);
    OSMetaClassDeclareReservedUnused(AnVMSR, 6);
    OSMetaClassDeclareReservedUnused(AnVMSR, 7);
};

class AnVMSRUserClient : public IOUserClient
{
	OSDeclareDefaultStructors(AnVMSRUserClient);

private:
    IOBufferMemoryDescriptor *memDesc = NULL;
	AnVMSR *mDevice = NULL;
    task_t fTask = NULL;
    bool haveCreatedRef = false;
    int Q_Err = 0;

#if USE_OLD_IODATAQUEUE != 0
    IODataQueue *myLogQueue = NULL;
#else /* USE_OLD_IODATAQUEUE == 0 */
    IOSharedDataQueue *myLogQueue = NULL;
#endif /* USE_OLD_IODATAQUEUE */

public:
    /* Message Handler Method */
    void messageHandler(UInt32 type, const char *format, ...) ATTRIBUTE_MACRO;

    /* Initialize User Client With Task Method */
    static const AnVMSRUserClient *withTask(task_t owningTask);

    /* IOKit Methods */
	virtual bool start(IOService *provider) APPLE_KEXT_OVERRIDE;
	virtual void stop(IOService *provider) APPLE_KEXT_OVERRIDE;
    virtual void free(void) APPLE_KEXT_OVERRIDE;
    virtual bool initWithTask(task_t owningTask, void * securityToken,
                              UInt32 type, OSDictionary * properties) APPLE_KEXT_OVERRIDE;
    virtual bool initWithTask(task_t owningTask, void * securityToken, UInt32 type) APPLE_KEXT_OVERRIDE;
	virtual IOReturn clientClose(void) APPLE_KEXT_OVERRIDE;
	virtual IOReturn clientDied(void) APPLE_KEXT_OVERRIDE;
    virtual bool willTerminate(IOService *provider, IOOptionBits options) APPLE_KEXT_OVERRIDE;
    virtual bool didTerminate(IOService *provider, IOOptionBits options, bool *defer) APPLE_KEXT_OVERRIDE;
    virtual bool terminate(IOOptionBits options = 0) APPLE_KEXT_OVERRIDE;
    virtual IOExternalMethod *getTargetAndMethodForIndex(IOService **targetP,
                                                         UInt32 index) APPLE_KEXT_OVERRIDE;
    virtual IOExternalAsyncMethod *getAsyncTargetAndMethodForIndex(IOService **target,
                                                                   UInt32 index) APPLE_KEXT_OVERRIDE;
    virtual IOReturn clientMemoryForType(UInt32 type, IOOptionBits *options,
                                         IOMemoryDescriptor **memory) APPLE_KEXT_OVERRIDE;

    /* Set Dictionary Data Methods */
    virtual bool MergeDictionaryIntoDictionary(OSDictionary *parentSourceDictionary,
                                               OSDictionary *parentTargetDictionary);
    virtual bool MergeDictionaryIntoProvider(IOService *provider, OSDictionary *dictionaryToMerge);

    /* Set Queue Size Method */
    virtual bool set_Q_Size(UInt64 capacity);

    /* Read Register Action Methods */
    virtual IOReturn actionMethodRDMSR(UInt32 *dataIn, UInt32 *dataOut,
                                       IOByteCount inputSize, IOByteCount *outputSize);         /* RDMSR */
    virtual IOReturn actionMethodRDPMC(UInt32 *dataIn, UInt32 *dataOut,
                                       IOByteCount inputSize, IOByteCount *outputSize);         /* RDPMC */
    virtual IOReturn actionMethodRDTSC(UInt32 *dataIn, UInt32 *dataOut,
                                       IOByteCount inputSize, IOByteCount *outputSize);         /* RDTSC */

    /* Read Register Async Action Methods */
    virtual IOReturn actionMethodAsyncRDMSR(OSAsyncReference __unused asyncRef,
                                            UInt32 *dataIn, UInt32 *dataOut,
                                            IOByteCount inputSize, IOByteCount *outputSize);    /* RDMSR */
    virtual IOReturn actionMethodAsyncRDPMC(OSAsyncReference __unused asyncRef,
                                            UInt32 *dataIn, UInt32 *dataOut,
                                            IOByteCount inputSize, IOByteCount *outputSize);    /* RDPMC */
    virtual IOReturn actionMethodAsyncRDTSC(OSAsyncReference __unused asyncRef,
                                            UInt32 *dataIn, UInt32 *dataOut,
                                            IOByteCount inputSize, IOByteCount *outputSize);    /* RDTSC */

    /* Write Register Action Methods */
    virtual IOReturn actionMethodWRMSR(UInt32 *dataIn, UInt32 *dataOut,
                                       IOByteCount inputSize, IOByteCount *outputSize);         /* WRMSR */
    virtual IOReturn actionMethodWRPMC(UInt32 *dataIn, UInt32 *dataOut,
                                       IOByteCount inputSize, IOByteCount *outputSize);         /* WRPMC */
    virtual IOReturn actionMethodWRTSC(UInt32 *dataIn, UInt32 *dataOut,
                                       IOByteCount inputSize, IOByteCount *outputSize);         /* WRTSC */

    /* Write Register Async Action Methods */
    virtual IOReturn actionMethodAsyncWRMSR(OSAsyncReference __unused asyncRef,
                                            UInt32 *dataIn, UInt32 *dataOut,
                                            IOByteCount inputSize, IOByteCount *outputSize);    /* WRMSR */
    virtual IOReturn actionMethodAsyncWRPMC(OSAsyncReference __unused asyncRef,
                                            UInt32 *dataIn, UInt32 *dataOut,
                                            IOByteCount inputSize, IOByteCount *outputSize);    /* WRPMC */
    virtual IOReturn actionMethodAsyncWRTSC(OSAsyncReference __unused asyncRef,
                                            UInt32 *dataIn, UInt32 *dataOut,
                                            IOByteCount inputSize, IOByteCount *outputSize);    /* WRTSC */

    /* Virtual Padding Functions For AnVMSRUserClient */
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 0);
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 1);
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 2);
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 3);
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 4);
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 5);
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 6);
    OSMetaClassDeclareReservedUnused(AnVMSRUserClient, 7);
};
